from django.conf import settings

DEBUG = getattr(settings, "DEBUG", False)
DJANGO_TABLES2_TEMPLATE = getattr(settings, "DJANGO_TABLES2_TEMPLATE", "django_tables2/bootstrap5.html")
DJBACKOFFICE_SHOW_USER_DJANGO = getattr(settings, "DJBACKOFFICE_SHOW_USER_DJANGO", True)



import sweetify
from django.forms.models import model_to_dict
from django.views.generic.detail import BaseDetailView, DetailView
from django.views.generic.edit import DeletionMixin
from django.views.generic.list import ListView
from django.views.generic.edit import CreateView, UpdateView
from django.urls import reverse_lazy
from djbackoffice.mixin import BackofficeContextMixin, BackOfficeFilterMixin
from djbackoffice.utils import Logger
from django_tables2 import SingleTableMixin
from django_tables2 import tables
from djbackoffice.tables import SimpleTable
from djbackoffice.settings import DJANGO_TABLES2_TEMPLATE


class ListView(BackofficeContextMixin, BackOfficeFilterMixin, SingleTableMixin, ListView):
    table_class = SimpleTable
    model_meta = None
    template_name = "djbackoffice/general_list.html"
    fields = ()
    crud_mode = ''

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        info = self.model_meta.app_label, self.model_meta.model_name
        if "c" in self.crud_mode.lower():
            context["add_url"] = "djbackoffice:%s_%s_add" % info
        else:
            context["add_url"] = None

        if "u" in self.crud_mode.lower():
            context["edit_url"] = "djbackoffice:%s_%s_edit" % info
        else:
            context["edit_url"] = None

        if "d" in self.crud_mode.lower():
            context["delete_url"] = "djbackoffice:%s_%s_delete" % info
        else:
            context["delete_url"] = None

        context["detail_url"] = "djbackoffice:%s_%s_detail" % info
        context["table_template"] = DJANGO_TABLES2_TEMPLATE
        return context

    def get_table_class(self):
        """
        override func from SingleTableMixin
        """
        fields = self.fields + ("actions",)
        return tables.table_factory(self.model, table=self.table_class, fields=fields)


class CreateView(BackofficeContextMixin, CreateView):
    model_meta = None
    fields = '__all__'
    form_colum_style = 1

    def form_valid(self, form):
        self.object = form.save()
        log = Logger()
        log.addition(self.request, self.object)
        return super().form_valid(form)

    def get_success_url(self):
        sweetify.toast(self.request, "Success add %s" % self.model_meta.model_name, timer=5000)
        info = self.model_meta.app_label, self.model_meta.model_name
        pattern = 'djbackoffice:%s_%s_list' % info
        return reverse_lazy(pattern)


class UpdateView(BackofficeContextMixin, UpdateView):
    model_meta = None
    fields = '__all__'
    form_colum_style = 1

    def form_valid(self, form):
        self.object = form.save()
        log = Logger()
        log.change(self.request, self.object)
        return super().form_valid(form)

    def get_success_url(self):
        sweetify.toast(self.request, "Success edit %s" % self.model_meta.model_name, timer=5000)
        info = self.model_meta.app_label, self.model_meta.model_name
        pattern = 'djbackoffice:%s_%s_list' % info
        return reverse_lazy(pattern)


class DeleteView(DeletionMixin, BaseDetailView):
    model_meta = None

    def delete(self, request, *args, **kwargs):
        log = Logger()
        obj = self.get_object()
        deletion = super().delete(request, *args, **kwargs)
        log.deletion(request, obj)
        return deletion

    def get(self, request, *args, **kwargs):
        return self.delete(request, *args, **kwargs)

    def get_success_url(self):
        sweetify.toast(self.request, "Success delete %s" % self.model_meta.model_name, timer=5000)
        info = self.model_meta.app_label, self.model_meta.model_name
        pattern = 'djbackoffice:%s_%s_list' % info
        return reverse_lazy(pattern)


class DetailView(BackofficeContextMixin, DetailView):
    template_name = "djbackoffice/general_detail.html"

    def get_title_page(self):
        obj = self.get_object()
        return f"Detail data {obj.id}"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        obj = self.get_object()
        context["object_dict"] = model_to_dict(obj)
        return context



__title__ = 'Django Backoffice Dashboard'
__name__ = 'django-backoffice-dashboard-alpha'
__version__ = '0.0.1.a0'
__author__ = 'irfanpule'


from django.utils.module_loading import autodiscover_modules
from djbackoffice.core import backoffice


def autodiscover():
    """
    autodiscover_modules all file backoffice.py to module
    """
    autodiscover_modules('backoffice', register_to=backoffice)



from django.db.models.base import ModelBase
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.text import capfirst
from django.apps import apps


class AlreadyRegistered(Exception):
    pass


class BackofficeOptionException(Exception):
    pass


class BackofficeOptions:
    """
    Class options to configure the register model
    """

    list_display = ()
    list_per_page = 50
    search_fields = ()
    form_column_style = 1

    # crud_mode to select func Create, Read, Update, Delete
    # default option is `crud`, and `r` or read mode can't disable
    crud_mode = 'crud'
    icon_menu = ''

    def __init__(self, model, opts_class):
        self.model = model
        self.model_meta = model._meta
        self.opts_class = opts_class
        super().__init__()

    def __str__(self):
        return "%s.%s" % (self.model_meta.app_label, self.__class__.__name__)

    def _get_field_names(self):
        """
        To get all field names of model
        """
        return [f.name for f in self.model._meta.get_fields()]

    def _get_name_model(self):
        return capfirst(self.model._meta.verbose_name_plural)

    def _get_name_app(self):
        return capfirst(apps.get_app_config(self.model._meta.app_label).verbose_name)

    def list_view(self, request):
        """
        CreateView to show all data of model
        """
        from djbackoffice.views import ListView
        class_view = ListView
        class_view.backoffice = self.opts_class
        class_view.model = self.model
        class_view.crud_mode = self.crud_mode
        class_view.search_fields = self.search_fields
        class_view.fields = self.list_display or tuple(self._get_field_names())
        class_view.paginate_by = self.list_per_page
        class_view.search_fields = self.search_fields
        class_view.model_meta = self.model_meta
        class_view.title_page = "List %s" % self._get_name_model()
        class_view.sub_title = "On App %s" % self._get_name_app()
        return class_view.as_view()(request)

    def add_view(self, request):
        """
        CreateView to add data of model
        """
        from djbackoffice.views import CreateView
        class_view = CreateView
        class_view.backoffice = self.opts_class
        class_view.model = self.model
        class_view.model_meta = self.model_meta
        class_view.title_page = "Add %s" % self._get_name_model()
        class_view.sub_title = "On App %s" % self._get_name_app()
        class_view.template_name = self.get_form_colum_style()
        return class_view.as_view()(request)

    def edit_view(self, request, pk):
        """
        UpdateView to update data of model
        """
        from djbackoffice.views import UpdateView
        class_view = UpdateView
        class_view.backoffice = self.opts_class
        class_view.model = self.model
        class_view.model_meta = self.model_meta
        class_view.title_page = "Edit %s" % self._get_name_model()
        class_view.sub_title = "On App %s" % self._get_name_app()
        class_view.template_name = self.get_form_colum_style()
        return class_view.as_view()(request, pk=pk)

    def delete_view(self, request, pk):
        """
        DeleteView to remove data of model
        """
        from djbackoffice.views import DeleteView
        class_view = DeleteView
        class_view.model = self.model
        class_view.model_meta = self.model_meta
        return class_view.as_view()(request, pk=pk)

    def detail_view(self, request, pk):
        """
        DetailView to show all data of model
        """
        from djbackoffice.views import DetailView
        class_view = DetailView
        class_view.backoffice = self.opts_class
        class_view.model = self.model
        class_view.model_meta = self.model_meta
        class_view.sub_title = "On App %s" % self._get_name_app()
        return class_view.as_view()(request, pk=pk)

    def get_urls(self):
        """
        Generates a CRUD url from a registered model
        """
        from django.urls import path
        info = self.model._meta.app_label, self.model._meta.model_name

        urlpatterns = [
            path('', self.list_view, name='%s_%s_list' % info),
            path('detail/<path:pk>/', self.detail_view, name='%s_%s_detail' % info),
        ]
        if 'c' in self.crud_mode.lower():
            print("create mode", self.model_meta)
            urlpatterns += [path('add/', self.add_view, name='%s_%s_add' % info)]
        if 'u' in self.crud_mode.lower():
            urlpatterns += [path('edit/<path:pk>/', self.edit_view, name='%s_%s_edit' % info)]
        if 'd' in self.crud_mode.lower():
            urlpatterns += [path('delete/<path:pk>/', self.delete_view, name='%s_%s_delete' % info)]

        return urlpatterns

    @property
    def urls(self):
        return self.get_urls()

    def get_form_colum_style(self):
        """
        Returns the template according to the desired number of columns
        """
        if self.form_column_style == 1:
            return 'djbackoffice/partials/one-column-form.html'
        elif self.form_column_style == 2:
            return 'djbackoffice/partials/two-column-form.html'
        else:
            raise BackofficeOptionException("form_column_style set using 1 or 2")


class Backoffice(object):
    """
    A Backoffice object encapsulates an instance of a backoffice. Models are
    registered with the Backoffice using the register() method.
    """
    index_title = 'Backoffice'

    def __init__(self, name='djbackoffice'):
        self._registry = {}
        self.name = name

    def register(self, model_or_iterable, opts_class=None):
        """
        Registers the given model(s) with the given BackofficeOptions.
        The model(s) should be Model classes, not instances.
        """
        if isinstance(model_or_iterable, ModelBase):
            model_or_iterable = [model_or_iterable]

        opts_class = opts_class or BackofficeOptions

        for model in model_or_iterable:
            if model in self._registry:
                if model in self._registry:
                    raise AlreadyRegistered('The model %s is already registered' % model.__name__)
            self._registry[model] = opts_class(model, self)

    def index(self, request):
        """
        Display the main backoffice index page
        """
        sidebar_menu = self.get_app_menu_register()

        context = {
            'title': self.index_title,
            'sidebar_menu': sidebar_menu,
        }

        request.current_app = self.name
        return TemplateResponse(request, 'djbackoffice/home.html', context)

    def get_urls(self):
        """
        Collect all urls of the registered models
        """
        from django.urls import path, include
        from django.contrib.admin import AdminSite

        # Admin-site views.
        admin = AdminSite()
        admin.login_template = 'djbackoffice/login.html'
        admin.logout_template = 'djbackoffice/login.html'

        urlpatterns = [
            path('', self.index, name='index'),
            path('login/', admin.login, name='login'),
            path('logout/', admin.logout, name='logout'),
        ]
        for model, backoffice_opts in self._registry.items():
            urlpatterns += [
                path('%s/%s/' % (model._meta.app_label, model._meta.model_name), include(backoffice_opts.urls)),
            ]
        return urlpatterns

    @property
    def urls(self):
        return self.get_urls(), 'djbackoffice', self.name

    def get_app_menu_register(self):
        """
        Creates a menu sidebar based on the app name and registered model
        """
        app_dict = {}

        models = self._registry
        for model, opts in models.items():
            app_label = model._meta.app_label

            info = (app_label, model._meta.model_name)
            model_dict = {
                'name': capfirst(model._meta.verbose_name_plural),
                'object_name': model._meta.object_name,
                'admin_url': reverse('djbackoffice:%s_%s_list' % info),
                'icon_menu': opts.icon_menu
            }

            if app_label in app_dict:
                app_dict[app_label]['models'].append(model_dict)
            else:
                app_dict[app_label] = {
                    'name': apps.get_app_config(app_label).verbose_name,
                    'app_label': app_label,
                    'models': [model_dict],
                }

        app_list = sorted(app_dict.values(), key=lambda x: x['name'].lower())

        for app in app_list:
            app['models'].sort(key=lambda x: x['name'])

        return app_list


backoffice = Backoffice()



from django.apps import AppConfig


class DjbackofficeConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'djbackoffice'

    def ready(self):
        super().ready()
        self.module.autodiscover()



# from django.db import models
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from djbackoffice.core import backoffice, BackofficeOptions
from djbackoffice.settings import DJBACKOFFICE_SHOW_USER_DJANGO


class UserBackofficeOpts(BackofficeOptions):
    list_display = ('username', 'first_name', 'last_name', 'email', 'is_superuser', 'is_staff', 'is_active')
    search_fields = ('username', 'first_name', 'last_name', 'email',)
    icon_menu = "bi-person-fill"


class GroupBackofficeOpts(BackofficeOptions):
    icon_menu = "bi-people-fill"


if DJBACKOFFICE_SHOW_USER_DJANGO:
    backoffice.register(get_user_model(), UserBackofficeOpts)
    backoffice.register(Group, GroupBackofficeOpts)



from django.forms.models import model_to_dict
from django.contrib.admin.models import DELETION, ADDITION, CHANGE, LogEntry
from django.contrib.contenttypes.models import ContentType


def get_content_type_for_model(obj):
    return ContentType.objects.get_for_model(obj, for_concrete_model=False)


class Logger:
    """
    Some of the methods below are inspired by the ModelAdmin class
    django.contrib.admin.options
    """
    def model_to_dict_str(self, obj):
        return str(model_to_dict(obj))

    def addition(self, request, obj, message=""):
        message = f"{message} | {self.model_to_dict_str(obj)}"

        return LogEntry.objects.log_action(
            user_id=request.user.pk,
            content_type_id=get_content_type_for_model(obj).pk,
            object_id=obj.pk,
            object_repr=str(obj),
            action_flag=ADDITION,
            change_message=message,
        )

    def change(self, request, obj, message=""):
        message = f"{message} | {self.model_to_dict_str(obj)}"

        return LogEntry.objects.log_action(
            user_id=request.user.pk,
            content_type_id=get_content_type_for_model(obj).pk,
            object_id=obj.pk,
            object_repr=str(obj),
            action_flag=CHANGE,
            change_message=message,
        )

    def deletion(self, request, obj, message=""):
        message = f"{message} | {self.model_to_dict_str(obj)}"

        return LogEntry.objects.log_action(
            user_id=request.user.pk,
            content_type_id=get_content_type_for_model(obj).pk,
            object_id=obj.pk,
            object_repr=str(obj),
            action_flag=DELETION,
            change_message=message
        )



import django_tables2 as tables


class SimpleTable(tables.Table):
    """
    Class to add action each of row
    """
    actions = tables.TemplateColumn(
        template_name="djbackoffice/partials/action_row_table.html"
    )



from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django.utils.text import unescape_string_literal, smart_split
from django.views.generic.base import ContextMixin
from djbackoffice.core import BackofficeOptions


class BackofficeContextMixin(ContextMixin):
    title_page = ''
    sub_title = ''
    btn_submit_name = ''
    backoffice: BackofficeOptions
    crud_mode = ''

    def get_title_page(self):
        return self.title_page

    def get_sub_title_page(self):
        return self.sub_title

    def get_btn_submit_name(self):
        return self.btn_submit_name

    def get_context_data(self, **kwargs):
        kwargs['title'] = self.get_title_page()
        kwargs['title_page'] = self.get_title_page()
        kwargs['sub_title_page'] = self.get_sub_title_page()
        kwargs['btn_submit_name'] = self.get_btn_submit_name()
        kwargs['sidebar_menu'] = self.get_sidebar_menu()
        return super().get_context_data(**kwargs)

    def get_sidebar_menu(self):
        sidebar_menu = self.backoffice.get_app_menu_register()
        return sidebar_menu


class BackOfficeFilterMixin:
    search_fields = ()
    term = ''

    def get_queryset(self):
        """Return queryset. Overriding from ListView"""
        queryset = super().get_queryset()
        if self.search_fields:
            self.term = self.request.GET.get('term', '')
            qs, search_use_distinct = self.get_search_results(self.model.objects.all(), self.term)
            if search_use_distinct:
                qs = qs.distinct()
            return qs
        else:
            return queryset

    def get_search_fields(self):
        """
        Return a sequence containing the fields to be searched whenever
        somebody submits a search query.
        """
        return self.search_fields

    def get_search_results(self, queryset, search_term):
        """
        This function inspiration from django.contrib.admin.ModelAdmin

        Return a tuple containing a queryset to implement the search
        and a boolean indicating if the results may contain duplicates.
        """
        # Apply keyword searches.
        def construct_search(field_name):
            if field_name.startswith("^"):
                return "%s__istartswith" % field_name[1:]
            elif field_name.startswith("="):
                return "%s__iexact" % field_name[1:]
            elif field_name.startswith("@"):
                return "%s__search" % field_name[1:]
            # Use field_name if it includes a lookup.
            opts = queryset.model._meta
            lookup_fields = field_name.split(LOOKUP_SEP)
            # Go through the fields, following all relations.
            prev_field = None
            for path_part in lookup_fields:
                if path_part == "pk":
                    path_part = opts.pk.name
                try:
                    field = opts.get_field(path_part)
                except FieldDoesNotExist:
                    # Use valid query lookups.
                    if prev_field and prev_field.get_lookup(path_part):
                        return field_name
                else:
                    prev_field = field
                    if hasattr(field, "get_path_info"):
                        # Update opts to follow the relation.
                        opts = field.get_path_info()[-1].to_opts
            # Otherwise, use the field with icontains.
            return "%s__icontains" % field_name

        may_have_duplicates = False
        search_fields = self.get_search_fields()
        if search_fields and search_term:
            orm_lookups = [
                construct_search(str(search_field)) for search_field in search_fields
            ]
            for bit in smart_split(search_term):
                if bit.startswith(('"', "'")) and bit[0] == bit[-1]:
                    bit = unescape_string_literal(bit)
                or_queries = models.Q(
                    *((orm_lookup, bit) for orm_lookup in orm_lookups),
                    _connector=models.Q.OR,
                )
                queryset = queryset.filter(or_queries)
            # TODO: temporary comment
            # hardcode may_have_duplicate set False
            # may_have_duplicates |= any(
            #     lookup_spawns_duplicates(self.backoffice, search_spec)
            #     for search_spec in orm_lookups
            # )
            may_have_duplicates = False
        return queryset, may_have_duplicates

    def get_context_data(self, **kwargs):
        kwargs['use_search_form'] = True if self.get_search_fields() else False
        kwargs['search_term'] = self.term
        return super().get_context_data(**kwargs)



def register(model_or_iterable, **options):
    """
    Registers the given model(s) with the given backoffice options.
    The model(s) should be Model classes, not instances.

    @register(Book)
    class BookBackoffice(BackofficeOptions):
        pass
    """
    from djbackoffice.core import BackofficeOptions, backoffice

    def wrapper(opts_class):
        if not issubclass(opts_class, BackofficeOptions):
            raise ValueError('Wrapped class must subclass BackofficeOptions.')
        backoffice.register(model_or_iterable, opts_class, **options)
        return opts_class

    return wrapper
